/****************************************************************************
Copyright (C) 2012 HidraVFX development team
            <https://launchpad.net/~hidravfx-dev-team>

This file is part of the HidraVFX project.

HidraVFX is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HidraVFX is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HidraVFX. If not, see <http://www.gnu.org/licenses/>.
****************************************************************************/

#include <math.h>
#include <mathplus.h>
#include "vertex.h"
#include "spline.h"


tcubicspline tcubicspline_of_tpoint(tpoint p0,tpoint p1,tpoint p2,tpoint p3)
{
  tcubicspline s;
  s.point[0] = tpoint_copy(&p0);
  s.point[1] = tpoint_copy(&p1);
  s.point[2] = tpoint_copy(&p2);
  s.point[3] = tpoint_copy(&p3);
  return s;
}

tcubicspline tcubicspline_of_double(double p0x,double p0y,double p1x,double p1y,double p2x,double p2y,double p3x,double p3y)
{
  tcubicspline s;
  s.point[0].x = p0x;
  s.point[0].y = p0y;
  s.point[0].z = 0.0;
  s.point[1].x = p1x;
  s.point[1].y = p1y;
  s.point[1].z = 0.0;
  s.point[2].x = p2x;
  s.point[2].y = p2y;
  s.point[2].z = 0.0;
  s.point[3].x = p3x;
  s.point[3].y = p3y;
  s.point[3].z = 0.0;
  return s;
}

/* 0 <== u <== 1 */
tpoint bspline_of_tpoint(tpoint p0, tpoint p1,tpoint p2, double u)
{
  tpoint p;
  double a1,a2,a3;

  a1 = u*u-2*u+1;
  a2 = -2*u*u+2*u+1;
  a3 = u*u;
  p.x = 0.5*(a1*p0.x + a2*p1.x + a3*p2.x);
  p.y = 0.5*(a1*p0.y + a2*p1.y + a3*p2.y);
  p.z = 0.0;
  return p;
}

/* 0 <== u <== 1 */
tpoint bspline_of_tcubicspline(tcubicspline p, double u)
{
  tpoint result;
  double a1,a2,a3,a4;

  a1 = - 1*u*u*u + 3*u*u - 3*u +1;
  a2 =   3*u*u*u - 6*u*u +     +4;
  a3 = - 3*u*u*u + 3*u*u + 3*u +1;
  a4 =   1*u*u*u;
  result.x = (1.0/6.0)*(a1*p.point[0].x + a2*p.point[1].x + a3*p.point[2].x + a4*p.point[3].x);
  result.y = (1.0/6.0)*(a1*p.point[0].y + a2*p.point[1].y + a3*p.point[2].y + a4*p.point[3].y);
  result.z = 0.0;
  return result;
}

/* catmull-rom curve in 2d */
/* u fut vltoz, rtke: [1] */
/* the curve is drawn between p1 & p2 */
/* p0 & p3 are extra interval control points */
tpoint catmullrom_of_tcubicspline(tcubicspline p, double u)
{
  tpoint result;
  double u2,u3;

  u2 = u * u;
  u3 = u2 * u;
  result.x = 0.5 * ( ( 2.0 * p.point[1].x ) +
                ( -p.point[0].x + p.point[2].x ) * u +
                ( 2.0 * p.point[0].x - 5.0 * p.point[1].x + 4.0 * p.point[2].x - p.point[3].x ) * u2 +
                ( -p.point[0].x + 3.0 * p.point[1].x - 3.0 * p.point[2].x + p.point[3].x ) * u3 );
   result.y = 0.5 * ( ( 2.0 * p.point[1].y ) +
              ( -p.point[0].y + p.point[2].y ) * u +
                ( 2.0 * p.point[0].y - 5.0 * p.point[1].y + 4 * p.point[2].y - p.point[3].y ) * u2 +
                ( -p.point[0].y + 3.0 * p.point[1].y - 3.0 * p.point[2].y + p.point[3].y ) * u3 );
   result.z = 0.0;
   return result;
}

/* bezier curve in 2d */
tpoint bezier_of_2tpoints(tpoint p0,tpoint p1, double u)
{
  tpoint result;

  result.x = (1.0 - u) * p0.x + u * p1.x;
  result.y = (1.0 - u) * p0.y + u * p1.y;
  result.z = 0.0;
  return result;
}

/* bezier curve in 2d */
/* u fut vltoz, rtke: [1] */
tpoint bezier_of_3tpoints(tpoint p0,tpoint p1,tpoint p2,double u)
{
  tpoint result;

  result.x = sqr(1-u) * p0.x + 2.0 * (1.0-u) * u * p1.x + sqr(u) * p2.x;
  result.y = sqr(1-u) * p0.y + 2.0 * (1.0-u) * u * p1.y + sqr(u) * p2.y;
  result.z = 0.0;
  return result;
}

/* bezier curve in 2d */
/* u fut vltoz, rtke: [1] */
tpoint bezier_of_tcubicspline(tcubicspline p,double u)
{
  tpoint result;

  result.x = sqr3(1.0-u)*p.point[0].x + 3.0*sqr(1.0-u)*u*p.point[1].x + 3.0*(1.0-u)*sqr(u)*p.point[2].x + sqr3(u)*p.point[3].x;
  result.y = sqr3(1.0-u)*p.point[0].y + 3.0*sqr(1.0-u)*u*p.point[1].y + 3.0*(1.0-u)*sqr(u)*p.point[2].y + sqr3(u)*p.point[3].y;
  result.z = 0.0;
  return result;
}

// first derivate of a bezier
tpoint beziertangent_of_tcubicspline(tcubicspline bezier, double u)
{
  tpoint result;

  result.x = - 3.0*sqr(1.0-u)*bezier.point[0].x + 3.0*sqr(1.0-u)*bezier.point[1].x
               - 6.0*(1.0-u)*u*bezier.point[1].x + 6.0*(1.0-u)*u*bezier.point[2].x
               - 3.0*u*u*bezier.point[2].x + 3.0*u*u*bezier.point[3].x;
  result.y = - 3.0*sqr(1.0-u)*bezier.point[0].y + 3.0*sqr(1.0-u)*bezier.point[1].y
               - 6.0*(1.0-u)*u*bezier.point[1].y + 6.0*(1.0-u)*u*bezier.point[2].y
               - 3.0*u*u*bezier.point[2].y + 3.0*u*u*bezier.point[3].y;
  result.z = 0.0;
  return result;
}

// second derivate of a bezier
tpoint bezierradius_of_tcubicspline(tcubicspline bezier, double u)
{
  tpoint result;

  result.x = 6.0*(1.0-u)*bezier.point[0].x + 6.0*u*bezier.point[3].x
               + (-6.0*(1.0-u)+6.0*u-6.0*(1.0-u))*bezier.point[1].x
               + (-6.0*u+6.0*(1.0-u)-6.0*u)*bezier.point[2].x;
  result.y = 6.0*(1.0-u)*bezier.point[0].y + 6.0*u*bezier.point[3].y
               + (-6.0*(1.0-u)+6.0*u-6.0*(1.0-u))*bezier.point[1].y
               + (-6.0*u+6.0*(1.0-u)-6.0*u)*bezier.point[2].y;
  result.z = 0.0;
  return result;
}
